#!/usr/bin/env python
# coding: utf-8

# In[1]:


class RtCamera:
    def __init__(self, position, direction, up, screen_distance=1, screen_width=1):
        '''Define the viewing position and direction, and the imaginary screen.'''
        self.position = position
        self.direction = direction
        self.up = up
        self.screen_distance = screen_distance
        self.screen_width = screen_width

    @property
    def viewing_angle(self):
        radians = 2 * math.atan2(self.screen_width / 2, self.screen_distance)
        return math.degrees(radians)

    @viewing_angle.setter
    def viewing_angle(self, value):
        if value < 1 or value > 180:
            raise ValueError(f'viewing_angle must be ' + \
                'between 1 and 179 degrees but has value {value}')
        radians = math.radians(value)
        self.screen_width = 2 * math.tan(radians / 2) * self.screen_distance


# In[2]:


from pygame import Color

class RtColor:
    '''Color components are between 0.0 and 1.0.'''
    def __init__(self, r = 0, g = 0, b = 0):
        self.r = r
        self.g = g
        self.b = b

    def __str__(self):
        return f'[{self.r}, {self.g}, {self.b}]'

    @staticmethod
    def from_bytes(r, g, b):
        '''Create an RtColor from bytes between 0 and 254.'''
        return RtColor(r / 255, g / 255, b / 255)

    @staticmethod
    def from_rgb_tuple(rgb):
        '''Create an RtColor from bytes between 0 and 254.'''
        return RtColor.from_bytes(rgb[0], rgb[1], rgb[2])

    @staticmethod
    def from_name(name):
        '''Create an RtColor from a color name.'''
        pygame_color = Color(name)
        return RtColor.from_bytes(pygame_color.r, pygame_color.g, pygame_color.b)

    def to_rgb_tuple(self):
        '''Convert the Color into an (R, G, B) tuple with values between 0 and 255.'''
        r = int(self.r * 255)
        g = int(self.g * 255)
        b = int(self.b * 255)
        return (r, g, b)

    def __mul__(self, other):
        '''Multiply an RtColor by another RtColor or a number.'''
        # See if other is a number.
        if isinstance(other, int) or isinstance(other, float):
            r = self.r * other
            g = self.g * other
            b = self.b * other
        else:
            r = self.r * other.r
            g = self.g * other.g
            b = self.b * other.b
        return RtColor(r, g, b)

    def __rmul__(self, other):
        '''Multiply another RtColor or a number by an RtColor.'''
        return self * other

    def __add__(self, other):
        '''
        Add corresponding components of two RtColors.
        For example, multiple lights add together.
        '''
        r = self.r + other.r
        g = self.g + other.g
        b = self.b + other.b
        return RtColor(r, g, b)

    def __sub__(self, other):
        '''Subtract corresponding components of two RtColors.'''
        r = self.r - other.r
        g = self.g - other.g
        b = self.b - other.b
        return RtColor(r, g, b)

# Define some useful colors.
RtColor.BLACK        = RtColor.from_name('black')
RtColor.WHITE        = RtColor.from_name('white')
RtColor.RED          = RtColor.from_name('red')
RtColor.GREEN        = RtColor.from_name('green')
RtColor.BLUE         = RtColor.from_name('blue')
RtColor.YELLOW       = RtColor.from_name('yellow')
RtColor.ORANGE       = RtColor.from_name('orange')
RtColor.PURPLE       = RtColor.from_name('purple')
RtColor.CYAN         = RtColor.from_name('cyan')
RtColor.MAGENTA      = RtColor.from_name('magenta')
RtColor.FOREST_GREEN = RtColor.from_name('forest green')
RtColor.LIGHT_GREEN  = RtColor.from_name('light green')
RtColor.DODGER_BLUE  = RtColor.from_name('dodger blue')
RtColor.ROYAL_BLUE   = RtColor.from_name('royal blue')
RtColor.LIGHT_BLUE   = RtColor.from_name('light blue')
RtColor.PINK         = RtColor.from_name('pink')
RtColor.DARK_GRAY    = RtColor.from_name('dark gray')
RtColor.GRAY         = RtColor.from_name('gray')
RtColor.LIGHT_GRAY   = RtColor.from_name('light gray')
RtColor.GRAY40       = RtColor.from_name('gray40')
RtColor.CHOCOLATE    = RtColor.from_name('chocolate')
RtColor.SADDLE_BROWN = RtColor.from_name('saddle brown')
RtColor.GOLD         = RtColor.from_name('gold')


# In[3]:


class RtLight:
    def __init__(self, color):
        self.color = color


# In[4]:


class RtAmbientLight(RtLight):
    def __init__(self, color, occlusion_weight=0,
            occlusion_range=0, num_occlusion_feelers=0):
        super().__init__(color)
        self.occlusion_weight = occlusion_weight
        self.occlusion_range = occlusion_range
        self.num_occlusion_feelers = num_occlusion_feelers

    def find_hit_color(self, scene, hit_info):
        '''Return the light's color multiplied by the material's ambient color.'''
        # Get the base color.
        if hit_info.object.material.texture is None:
            base_color = self.color * hit_info.object.material.ambient_color
        else:
            base_color = self.color * hit_info.object.color_at_point(hit_info)

        # If we are not using ambient occlusion feelers,
        # simply return the color.
        if self.num_occlusion_feelers == 0: return base_color

        # Perform ambient occlusion.
        # Cast ambient occlusion feelers and see how many are blocked.
        shadow_amount = 0
        for i in range(self.num_occlusion_feelers):
            feeler = random_directed_vector(hit_info.normal)
            feeler_hit_info = scene.find_hit(hit_info.point, feeler)

            # If there is a blocking object, add (occlusion_range - t)
            # so objects closer to the point add more shadow.
            if feeler_hit_info.object is not None:
                new_amount = self.occlusion_range - feeler_hit_info.t
                if new_amount > 0:
                    shadow_amount += new_amount

        # Get the fraction shadowed.
        fraction_shadowed = shadow_amount / (self.num_occlusion_feelers * self.occlusion_range)

        # Get the fraction not shadowed.
        fraction_clear = 1.0 - fraction_shadowed

        # Return a weighted average of the color and the occluded color.
        return base_color * (self.occlusion_weight * fraction_clear) + \
            base_color * (1 - self.occlusion_weight)


# In[5]:


import math

class RtDirectionalLight(RtLight):
    def __init__(self, color, direction):
        super().__init__(color)
        self.direction = direction.normalize()

    def ray_is_blocked(self, scene, hit_point, light_direction, max_t):
        '''Return True if the shadow feeler is blocked.'''
        feeler_hit_info = scene.find_hit(hit_point, light_direction)
        return feeler_hit_info.t < max_t

    def find_ray_hit_color(self, direction_to_light, hit_info, attenuation):
        '''Return the color at a point for a light that uses rays.'''
        # Start with black.
        total_color = RtColor()

        ########################
        # Diffuse contribution #
        ########################
        # Get cosine(theta) where theta is the angle between
        # the normal and the direction to the light.
        cos_theta = direction_to_light.dot(hit_info.normal)

        # Add a diffuse contribution by using Lambert's cosine law:
        #      Id = Ii * kd * Cos(theta)     for 0 <= theta <= Pi/2
        # If Cos(theta) < 0, this is a backface so make no color contribution.
        if cos_theta > 0:
            if hit_info.object.material.texture is None:
                total_color += \
                    (self.color * hit_info.object.material.diffuse_color) * (cos_theta * attenuation)
            else:
                total_color += \
                    (self.color * hit_info.object.color_at_point(hit_info)) * (cos_theta * attenuation)

        #########################
        # Specular contribution #
        #########################
        # Get cos(beta) where beta is the angle between
        # ray_direction and the direction to the light.
        mirror_vector = find_mirror_vector(-direction_to_light, hit_info.normal)
        cos_beta = -mirror_vector.dot(hit_info.ray_direction)

        # Add a specular contribution by using Phong's formula:
        #      Is = Is * ks * Cos(beta)^spec_n     for 0 <= beta <= Pi/2
        # If cos(beta) < 0, this is a backface so make no color contribution.

        # Calculate the specular component.
        if cos_beta > 0:
            phong_value = cos_beta ** hit_info.object.material.specular_power
            intensity = hit_info.object.material.specular_intensity
            total_color += self.color * (intensity * phong_value * attenuation)

        # Return the total color.
        return total_color

    def find_hit_color(self, scene, hit_info):
        '''Return the hit point's color.'''
        # See if the light is blocked at any distance.
        if self.ray_is_blocked(scene, hit_info.point, -self.direction, max_t=math.inf):
            return RtColor.BLACK

        # Calculate the color.
        # For directional lights, attenuation = 1.0 so there is no intensity reduction.
        return self.find_ray_hit_color(-self.direction, hit_info, attenuation=1.0)


# In[6]:


import math

class RtPointLight(RtDirectionalLight):
    def __init__(self, color, location, range=math.inf,
            constant_attenuation=0.0,
            linear_attenuation=0.0,
            quadratic_attenuation=1.0):
        self.color = color
        self.location = location
        self.range = range
        self.constant_attenuation = constant_attenuation
        self.linear_attenuation = linear_attenuation
        self.quadratic_attenuation = quadratic_attenuation

    def find_hit_color(self, scene, hit_info):
        '''Return the hit point's color.'''
        # If the hit point is out of the light's range, return black.
        ray_to_light = self.location - hit_info.point
        if ray_to_light.length() > self.range: return RtColor.BLACK

        # If the light is blocked, return black.
        if self.ray_is_blocked(scene, hit_info.point, ray_to_light, 1.0):
            return RtColor.BLACK

        # Calculate attenuation.
        ray_length = ray_to_light.length()
        attenuation = 1.0 / (
            self.constant_attenuation +
            ray_length * self.linear_attenuation +
            ray_length * ray_length * self.quadratic_attenuation
        )

        # Calculate the color.
        ray_to_light = ray_to_light.normalize()
        return self.find_ray_hit_color(ray_to_light, hit_info, attenuation)


# In[7]:


import math

class RtSpotLight(RtPointLight):
    def __init__(self, color, location, direction,
            inner_cone_angle=15,
            outer_cone_angle=30,
            range=math.inf,
            constant_attenuation=0.0,
            linear_attenuation=0.0,
            quadratic_attenuation=1.0):
        super().__init__(color, location, range,
            constant_attenuation, linear_attenuation, quadratic_attenuation)
        self.direction = direction.normalize()
        self.inner_cone_angle = inner_cone_angle
        self.outer_cone_angle = outer_cone_angle

    def find_hit_color(self, scene, hit_info):
        '''Return the hit point's color.'''
        # If the hit point is out of the light's range, return black.
        ray_to_light = self.location - hit_info.point
        if ray_to_light.length() > self.range: return RtColor.BLACK

        # If the light is blocked, return black.
        if self.ray_is_blocked(scene, hit_info.point, ray_to_light, 1.0):
            return RtColor.BLACK

        # Calculate base attenuation.
        ray_length = ray_to_light.length()
        attenuation = 1.0 / (
            self.constant_attenuation +
            ray_length * self.linear_attenuation +
            ray_length * ray_length * self.quadratic_attenuation
        )

        # Calculate the ray's cone angle.
        ray_to_light = ray_to_light.normalize()
        cosine = self.direction.dot(-ray_to_light)
        cone_angle = math.degrees(math.acos(cosine))
        if cone_angle > self.outer_cone_angle:
            return RtColor.BLACK

        # Attenuate across the cone angle range.
        if cone_angle > self.inner_cone_angle:
            # Calculate the fraction of the way across the range.
            fraction = (self.outer_cone_angle - cone_angle) / (self.outer_cone_angle - self.inner_cone_angle)
            attenuation *= fraction

        # Calculate the color.
        ray_to_light = ray_to_light.normalize()
        return self.find_ray_hit_color(ray_to_light, hit_info, attenuation)


# In[8]:


import math

class RtHitInfo:
    '''Save information about a hit point.'''
    def __init__(self, object=None, t=math.inf, point=None, normal=None, ray_direction=None, coords=None):
        self.object = object
        self.t = t
        self.point = point
        self.normal = normal
        self.ray_direction = ray_direction
        self.coords = coords

    def __str__(self):
        return f'{{{self.object}, {self.t}}}'

    def is_miss(self):
        return math.isinf(self.t)

    def is_hit(self):
        return not math.isinf(self.t)


# In[9]:


from PIL import Image

class RtMaterial:
    def __init__(self, ambient_color, diffuse_color=None,
            specular_intensity=0, specular_power=10, reflectivity=0,
            refractive_intensity=0, refractive_index=1):
        self.ambient_color = ambient_color
        if diffuse_color is None:
            diffuse_color = ambient_color
        self.diffuse_color = diffuse_color
        self.specular_intensity = specular_intensity
        self.specular_power = specular_power
        self.reflectivity = reflectivity
        self.refractive_intensity = refractive_intensity
        self.refractive_index = refractive_index
        self.texture = None
        self.bump_map = None
        self.normal_map = None

    def load_texture(self, filename):
        '''Load a texture image.'''
        image = Image.open(filename)
        self.texture_size = image.size  # Save the texture's size.
        self.texture = image.load()     # Load the texture's pixels.

    def load_bump_map(self, filename, bump_map_scale):
        '''Load a bump map image.'''
        image = Image.open(filename)
        self.bump_map_size = image.size  # Save the bump map's size.
        self.bump_map = image.load()     # Load the bump map's pixels.
        self.bump_map_scale = bump_map_scale

    def load_normal_map(self, filename):
        '''Load a normal map image.'''
        image = Image.open(filename)
        self.normal_map_size = image.size  # Save the normal map's size.
        self.normal_map = image.load()     # Load the normal map's pixels.

    def bump_map_height(self, u, v):
        # Add the color components.
        color = self.bump_map[u, v]
        total = color[0] + color[1] + color[2]

        # Scale from -1 to +1.
        half = 3 * 255 / 2
        total = (total - half) / half

        # Multiply by the bump map scale.
        return total * self.bump_map_scale


# In[10]:


# Mathy cell 1.
import math
import random

'''Intersection methods return RtHitInfo objects representing a point of intersection.'''

def intersect_ray_sphere(sphere, ray_origin, ray_direction):
    # Get the quadratic equation coefficients.
    a = ray_direction.dot(ray_direction)
    d = ray_origin - sphere.center
    b = 2 * ray_direction.dot(d)
    c = d.dot(d) - sphere.radius * sphere.radius

    # Check the discriminant.
    discriminant = b * b - 4 * a * c

    # No real intersection.
    if discriminant < 0: return RtHitInfo(sphere)
    discriminant = math.sqrt(discriminant)

    # If t < TINY, treat it as if it is behind the ray's origin.
    TINY = 1e-6

    # Find the solutions.
    t1 = (-b + discriminant) / (2 * a)
    if t1 < TINY: t1 = math.inf

    t2 = (-b - discriminant) / (2 * a)
    if t2 < TINY: t2 = math.inf

    # Find the smaller T value.
    if t2 < t1: t1 = t2

    # If we have no hit, return the default miss information.
    if math.isinf(t1): return RtHitInfo(sphere)

    # Return the hit information.
    point = ray_origin + t1 * ray_direction
    normal = (point - sphere.center).normalize()
    return RtHitInfo(sphere, t1, point, normal, ray_direction)

def find_mirror_vector(vector, normal):
    '''Find a vector's mirror vector.'''
    return vector - 2 * normal * vector.dot(normal)

def find_refraction_vector(I, N, ni, nt):
    '''Find a vector's refraction vector.'''
    # See: Wikipedia article "Snell's law" at tinyurl.com/mlhcu5v
    # See: Physics Stack Exchange question "What does the equation
    #      of a refracted ray trace mean?" at tinyurl.com/yclubtfy.

    # Use the dot product to calculate cos_i.
    # Rounding error sometimes makes cos_i something like +/-1.0000000000000002.
    cos_i = I.dot(N)
    if cos_i > 1: cos_i = 1
    if cos_i < -1: cos_i = -1

    # See if we're entering or leaving the object.
    if cos_i < 0:
        # Entering. Make cos_i positive.
        cos_i = -cos_i
    else:
        # Exiting. Swap ni and nt.
        ni, nt = nt, ni

    nr = ni / nt
    sin_i = math.sqrt(1 - cos_i * cos_i)
    sin_t = nr * sin_i

    # If sin_t >= 1, then we are at or beyond the critical angle.
    # In that case, return a null vector.
    if sin_t >= 1: return None

    cos_t = math.sqrt(1 - sin_t * sin_t)

    vector = nr * I + (nr * cos_i - cos_t) * N
    return vector.normalize()

def intersect_ray_plane(plane, ray_origin, ray_direction):
    # Get the dot product between the normal and ray direction.
    denominator = ray_direction.dot(plane.normal)

    # If the denominator is 0, then the ray is parallel
    # to the plane so there is no interesection.
    TINY = 1e-6
    if abs(denominator) < TINY: return RtHitInfo(plane)

    # Find t.
    w = ray_origin - plane.origin
    t = -plane.normal.dot(w) / denominator

    # If t < TINY, the intersection is behind the ray's
    # origin so there is no interesting intersection.
    if t < TINY: return RtHitInfo(plane)

    # Return the hit information.
    point = ray_origin + t * ray_direction
    return RtHitInfo(plane, t, point, plane.normal, ray_direction)

def gaussian_elimination(values):
    '''
    Use Gaussian elimination to solve a system of equations of the form:
      | A1 B1 ... N1 C1 |
      | A2 B2 ... N2 C2 |
      |       ...       |
      | Am Bm ... Nm Cm |
    The values parameter is a list of lists.
    Return the results in the final column.
    Return true if successful, false if there is no solution.
    '''
    TINY = 0.001
    num_rows = len(values)
    num_cols = len(values[0])
    for c in range(num_cols - 1):
        # Process column c.
        # Find a row to use.
        for r in range(c, num_rows):
            if abs(values[r][c]) > TINY:
                # Use this row. Swap it with row c.
                if r != c:
                    values[r], values[c] = values[c], values[r]
                break

        # If there's no usable row, then there's no solution.
        if abs(values[c][c]) <= TINY:
            return False

        # Change the leading coefficient in row c to 1
        coeff = values[c][c]
        for i in range(c, num_cols):
            values[c][i] /= coeff

        # Process the other rows by using column c.
        for r in range(num_rows):
            if r != c:
                # Subtract (row c) * values[r, c] from (row r).
                factor = values[r][c]
                for i in range(c, num_cols):
                    values[r][i] -= factor * values[c][i]

    # Return a success code.
    return True

def intersect_ray_vector_plane(ray_origin, ray_direction, a, ab, ac):
    # Make an array for Gaussian elimination.
    a0 = ray_origin - a
    values = [
        [ray_direction.x, -ab.x, -ac.x, -a0.x],
        [ray_direction.y, -ab.y, -ac.y, -a0.y],
        [ray_direction.z, -ab.z, -ac.z, -a0.z],        
    ]
    if not gaussian_elimination(values):
        t = math.inf
        u = 0
        v = 0
    else:
        t = values[0][3]
        u = values[1][3]
        v = values[2][3]
    return t, u, v

def intersect_ray_parallelogram(parallelogram, ray_origin, ray_direction):
    # Find the intersection.
    t, u, v = intersect_ray_vector_plane(ray_origin, ray_direction,
        parallelogram.a, parallelogram.ab, parallelogram.ac)
    if math.isinf(t): return RtHitInfo(parallelogram)

    TINY = 1e-6
    if (t < TINY) or (u < 0) or (v < 0) or (u > 1) or (v > 1):
        return RtHitInfo(parallelogram)

    point = ray_origin + t * ray_direction
    return RtHitInfo(parallelogram, t, point, parallelogram.normal, ray_direction)

def intersect_ray_triangle(triangle, ray_origin, ray_direction):
    # Find the intersection.
    t, u, v = intersect_ray_vector_plane(ray_origin, ray_direction,
        triangle.a, triangle.ab, triangle.ac)
    if math.isinf(t): return RtHitInfo(triangle)

    TINY = 1e-6
    if (t < TINY) or (u < 0) or (v < 0) or (u + v > 1):
        return RtHitInfo(triangle)

    point = ray_origin + t * ray_direction
    return RtHitInfo(triangle, t, point, triangle.normal, ray_direction)

def intersect_ray_checkerboard(checkerboard, ray_origin, ray_direction):
    # Find the intersection.
    t, u, v = intersect_ray_vector_plane(ray_origin, ray_direction,
        checkerboard.a, checkerboard.ab, checkerboard.ac)
    if math.isinf(t): return RtHitInfo(checkerboard)

    # See if it's in the general area.
    TINY = 1e-6
    if (t < TINY) or (u < 0) or (v < 0) or (u > checkerboard.num_ab) or (v > checkerboard.num_ac):
        return RtHitInfo(checkerboard)

    # See if it's on a used section.
    row = int(u)
    col = int(v)
    if (row + col) % 2 != checkerboard.parity: return RtHitInfo(checkerboard)

    point = ray_origin + t * ray_direction
    return RtHitInfo(checkerboard, t, point, checkerboard.normal, ray_direction)

def intersect_ray_mesh_triangle(mesh, triangle_num, ray_origin, ray_direction):
    # Find the intersection.
    # Get the triangle's first vertex and vectors.
    a = mesh.vertices[mesh.triangle_indices[triangle_num][0]]
    ab = mesh.ab_vectors[triangle_num]
    ac = mesh.ac_vectors[triangle_num]

    # Find the intersection.
    t, u, v = intersect_ray_vector_plane(ray_origin, ray_direction, a, ab, ac)
    if math.isinf(t): return RtHitInfo(mesh)

    TINY = 1e-6
    if (t < TINY) or (u < 0) or (v < 0) or (u + v > 1):
        return RtHitInfo(mesh)

    # We found an intersection. Find the point of intersection.
    point = ray_origin + t * ray_direction

    # Find the normal.
    if mesh.is_smooth:
        normal = mesh.find_smooth_normal(point, triangle_num, v)
    else:
        normal = mesh.triangle_normals[triangle_num]

    # Find the texture coordinates.
    if mesh.material.texture is None:
        coords = None
    else:
        index_a = mesh.triangle_indices[triangle_num][0]
        index_b = mesh.triangle_indices[triangle_num][1]
        index_c = mesh.triangle_indices[triangle_num][2]
        point_a = mesh.vertices[index_a]
        point_b = mesh.vertices[index_b]
        point_c = mesh.vertices[index_c]
        coords_a = mesh.texture_coords[index_a]
        coords_b = mesh.texture_coords[index_b]
        coords_c = mesh.texture_coords[index_c]
        coords = interpolate(point, point_a, point_b, point_c, coords_a, coords_b, coords_c, v)

    # Return the RtHitInfo.
    return RtHitInfo(mesh, t, point, normal, ray_direction, coords)

def map_range(value, from_min, from_max, to_min, to_max):
    return (value - from_min) / (from_max - from_min) * (to_max - to_min) + to_min

###############################################################
# Use one of the following versions of intersect_ray_cylinder #
###############################################################

# Based on Chris Dragan's calculations:
# https://hugi.scene.org/online/hugi24/coding%20graphics%20chris%20dragan%20raytracing%20shapes.htm
def intersect_ray_cylinder(cylinder, ray_origin, ray_direction):
    '''Find the intersections between the ray and cylinder.'''
    # Define some variables to make the calculations neater.
    D = ray_direction
    V = cylinder.axis
    r = cylinder.radius
    r = cylinder.radius
    X = ray_origin - cylinder.origin

    # Get the quadratic equation coefficients.
    a = D.dot(D) - D.dot(V) ** 2

    b_over_2 = D.dot(X) - D.dot(V) * X.dot(V)
    b = 2 * b_over_2

    c = X.dot(X) - X.dot(V) ** 2 - r ** 2

    # Check the discriminant.
    discriminant = b * b - 4 * a * c
    if discriminant < 0: return RtHitInfo(cylinder)
    discriminant = math.sqrt(discriminant)

    # Find the solutions.
    t1 = (-b + discriminant) / (2 * a)
    t2 = (-b - discriminant) / (2 * a)

    # Only work with t values >= 0.
    TINY = 1e-6
    if t1 < TINY:
        t1 = math.inf
    else:
        # Calculate the POI.
        intersection1 = ray_origin + t1 * ray_direction
        height1 = (intersection1 - cylinder.origin).dot(cylinder.axis)
        if height1 < 0 or height1 > cylinder.length: t1 = math.inf

    if t2 < TINY:
        t2 = math.inf
    else:
        # Calculate the POI.
        intersection2 = ray_origin + t2 * ray_direction
        height2 = (intersection2 - cylinder.origin).dot(cylinder.axis)
        if height2 < 0 or height2 > cylinder.length: t2 = math.inf

    # Find the smaller T value.
    if t2 < t1: t1 = t2

    point = ray_origin + t1 * ray_direction
    return RtHitInfo(cylinder, t1, point, cylinder.find_normal(point), ray_direction)

# Based on Denis Zorin's calculations.
# https://mrl.cs.nyu.edu/~dzorin/rend05/lecture2.pdf
def intersect_ray_cylinder(cylinder, ray_origin, ray_direction):
    '''Find the intersections between the ray and cylinder.'''
    # Define some variables to make the calculations neater.
    p = ray_origin
    v = ray_direction
    pa = cylinder.origin
    va = cylinder.axis
    dp = p - pa
    r = cylinder.radius

    # Get the quadratic equation coefficients.
    a1 = v - v.dot(va) * va
    a = a1.dot(a1)

    b = 2 * ((v - v.dot(va) * va).dot(dp - dp.dot(va) * va))

    c1 = dp - dp.dot(va) * va
    c = c1.dot(c1) - r ** 2

    # Check the discriminant.
    discriminant = b * b - 4 * a * c
    if discriminant < 0: return RtHitInfo(cylinder)
    discriminant = math.sqrt(discriminant)

    # Find the solutions.
    t1 = (-b + discriminant) / (2 * a)
    t2 = (-b - discriminant) / (2 * a)

    # Only work with t values >= 0.
    TINY = 1e-6
    if t1 < TINY:
        t1 = math.inf
    else:
        # See if the point is on the finite cylinder.
        intersection1 = ray_origin + t1 * ray_direction
        L = (intersection1 - cylinder.origin).dot(cylinder.axis)
        if L < 0 or L > cylinder.length: t1 = math.inf

    if t2 < TINY:
        t2 = math.inf
    else:
        # See if the point is on the finite cylinder.
        intersection2 = ray_origin + t2 * ray_direction
        L = (intersection2 - cylinder.origin).dot(cylinder.axis)
        if L < 0 or L > cylinder.length: t2 = math.inf

    # Find the smaller T value.
    if t2 < t1: t1 = t2

    point = ray_origin + t1 * ray_direction
    return RtHitInfo(cylinder, t1, point, cylinder.find_normal(point), ray_direction)

def interpolate(point, point_a, point_b, point_c, value_a, value_b, value_c, v):
    '''Use bilinear interpolation to find a v value for the point.'''
    # Find the p1 and p2 points.
    ac = point_c - point_a
    bc = point_c - point_b
    p1 = point_a + v * ac
    p2 = point_b + v * bc
    value_1 = (1 - v) * value_a + v * value_c
    value_2 = (1 - v) * value_b + v * value_c

    # Find the target point.
    p1_point = point - p1
    p1_p2 = p2 - p1
    w = p1_point.length() / p1_p2.length()
    return (1 - w) * value_1 + w * value_2

# See https://tinyurl.com/5ctmer4n
def cartesian_to_cylindrical(point):
    '''Convert from 3-D cartesian coordinates to 2-D cylindrical coordinates.'''
    # Convert from left-handed mathematical coordinates.
    x = point.z
    y = point.x
    z = -point.y

    r = math.sqrt(x * x + y * y)
    theta = math.atan2(y, x)
    return r, theta, z

def interpolate_rectangle(x, y, xmin, xmax, ymin, ymax,
        nw_value, ne_value, sw_value, se_value):
    '''Use bilinear interpolation to find a value for the point in a 2D rectangle.'''
    # Find the values at p1 and p2.
    v1 = nw_value + (sw_value - nw_value) * ((y - ymin) / (ymax - ymin))
    v2 = ne_value + (se_value - ne_value) * ((y - ymin) / (ymax - ymin))

    # Find the value at the point.
    return v1 + (v2 - v1) * ((x - xmin) / (xmax - xmin))

def interpolate_color(texture, texture_size, x, y):
    global interpolate_texture
    if not interpolate_texture:
        return RtColor.from_rgb_tuple(texture[x, y])
    else:
        xmin = int(x)
        xmax = xmin + 1
        if xmax >= texture_size[0]: xmax = texture_size[0] - 1

        ymin = int(y)
        ymax = ymin + 1
        if ymax >= texture_size[1]: ymax = texture_size[1] - 1

        nw_value = RtColor.from_rgb_tuple(texture[xmin, ymin])
        ne_value = RtColor.from_rgb_tuple(texture[xmax, ymin])
        sw_value = RtColor.from_rgb_tuple(texture[xmin, ymax])
        se_value = RtColor.from_rgb_tuple(texture[xmax, ymax])

        return interpolate_rectangle(x, y, xmin, xmax, ymin, ymax,
            nw_value, ne_value, sw_value, se_value)

def random_vector():
    '''Return a unit vector pointing in a random direction.'''
    while True:
        # Pick a random point inside the +/-1 cube.
        vector = Vector3(
            random.uniform(-1, 1),
            random.uniform(-1, 1),
            random.uniform(-1, 1))

        # If the distance to the origin is between TINY and 1, return it.
        TINY = 0.01
        if TINY <= vector.length() <= 1.0:
            return vector.normalize()

def random_directed_vector(direction):
    '''
    Return a unit vector pointing in a random direction
    in the hemisphere determined by vector direction.
    We assume that direction has length 1.
    '''
    # Get a random vector pointing in any direction.
    vector = random_vector()

    # If the vector is not pointing in the general
    # direction of the direction vector, reverse it.
    if vector.dot(direction) < 0:
        vector = -vector
    return vector

def random_disk_vector(vx, vy, max_length):
    '''
    Return a random vector in the plane defined by
    vectors vx and vy and with length at most max_length.
    '''
    while True:
        # Pick a random point inside the square +/-max_length.
        point = Vector2(
            random.uniform(-max_length, max_length),
            random.uniform(-max_length, max_length))

        # If the vector's length <= max_length, use it.
        if point.length() <= max_length:
            return vx * point.x + vy * point.y

def cartesian_to_spherical(point):
    '''Convert from cartesian to spherical coorsinates.'''
    # rho >= 0
    # -pi <= theta <= pi
    # 0 <= phi <= pi
    x, y, z = point.x, point.y, point.z

    rho = math.sqrt(x * x + y * y + z * z)
    TINY = 1e-6
    if rho < TINY:
        theta = 0
        phi = 0
    else:
        r = math.sqrt(x * x + z * z)
        if r < TINY:
            theta = 0
        else:
            theta = math.atan2(x, z)
        phi = math.acos(y / rho)
    return Vector3(rho, theta, phi)

def spherical_to_cartesian(point):
    rho, theta, phi = point.x, point.y, point.z
    y = rho * math.cos(phi)
    r = rho * math.sin(phi)
    x = r * math.sin(theta)
    z = r * math.cos(theta)
    return Vector3(x, y, z)


# In[11]:


# Mathy cell 2.
import math
from pygame import Vector2

# This cell holds methods that generate points for shapes like polygons and the Platonic solids.

def make_polygon_points(center, v1, v2, num_points):
    '''Return a list of points to make a polygon centered at center and in the plane defined by vectors v1 and v2.'''
    theta = 0
    dtheta = 2 * math.pi / num_points
    points = []
    for i in range(num_points):
        point = center + v1 * math.cos(theta) + v2 * math.sin(theta)
        points.append(point)
        theta += dtheta
    return points

# Methods that return vertices for the Platonic solids.
# Each method returns a tuple of vertices for a solid centered at
# the indicated center point and with the given circumradius.

def get_tetrahedron_points(center, circumradius):
    '''Return a tuple of tetrahedron vertices.'''
    dy = 0.25 * math.sqrt(2.0 / 3.0)

    tetrahedron_circumradius = math.sqrt(2.0 / 3.0) * 0.75
    scale = circumradius / tetrahedron_circumradius
    a = center + scale * Vector3(0, math.sqrt(2.0 / 3.0) - dy, 0)
    b = center + scale * Vector3(1.0 / math.sqrt(3.0), -dy, 0)
    c = center + scale * Vector3(-1.0 / (2 * math.sqrt(3.0)), -dy, -1.0 / 2.0)
    d = center + scale * Vector3(-1.0 / (2 * math.sqrt(3.0)), -dy, 1.0 / 2.0)
    return (a, b, c, d)

def get_tetrahedron_faces():
    '''Return a tuple holding tuples that define the solid's faces.'''
    return (
        (0, 1, 2),
        (0, 2, 3),
        (0, 3, 1),
        (3, 2, 1),
    )

def get_cube_points(center, circumradius):
    '''Return a tuple of cube vertices.'''
    cube_circumradius = math.sqrt(3.0)
    scale = circumradius / cube_circumradius
    a = center + scale * Vector3(-1, +1, +1)
    b = center + scale * Vector3(+1, +1, +1)
    c = center + scale * Vector3(+1, +1, -1)
    d = center + scale * Vector3(-1, +1, -1)
    e = center + scale * Vector3(-1, -1, +1)
    f = center + scale * Vector3(+1, -1, +1)
    g = center + scale * Vector3(+1, -1, -1)
    h = center + scale * Vector3(-1, -1, -1)
    return (a, b, c, d, e, f, g, h)

def get_cube_faces():
    '''Return a tuple holding tuples that define the cube's faces.'''
    return (
        (0, 1, 2, 3),
        (0, 3, 7, 4),
        (0, 4, 5, 1),
        (6, 2, 1, 5),
        (6, 5, 4, 7),
        (6, 7, 3, 2),
    )

def get_cube_texture_coords():
    '''Return a tuple holding tuples that define the cube's texture coordinates.'''
    return (
        ( Vector2(0.00, 0.00), Vector2(0.00, 0.50), Vector2(0.33, 0.50), Vector2(0.33, 0.00) ),
        ( Vector2(0.33, 0.00), Vector2(0.33, 0.50), Vector2(0.67, 0.50), Vector2(0.67, 0.00) ),
        ( Vector2(0.67, 0.00), Vector2(0.67, 0.50), Vector2(1.00, 0.50), Vector2(1.00, 0.00) ),
        ( Vector2(0.00, 0.50), Vector2(0.00, 1.00), Vector2(0.33, 1.00), Vector2(0.33, 0.50) ),
        ( Vector2(0.33, 0.50), Vector2(0.33, 1.00), Vector2(0.67, 1.00), Vector2(0.67, 0.50) ),
        ( Vector2(0.67, 0.50), Vector2(0.67, 1.00), Vector2(1.00, 1.00), Vector2(1.00, 0.50) ),
    )

def get_octahedron_points(center, circumradius):
    '''Return a tuple of octahedron vertices.'''
    octahedron_circumradius = 1
    scale = circumradius / octahedron_circumradius
    a = center + scale * Vector3(0, 1, 0)
    b = center + scale * Vector3(1, 0, 0)
    c = center + scale * Vector3(0, 0, -1)
    d = center + scale * Vector3(-1, 0, 0)
    e = center + scale * Vector3(0, 0, 1)
    f = center + scale * Vector3(0, -1, 0)
    return (a, b, c, d, e, f)

def get_octahedron_faces():
    '''Return a tuple holding tuples that define the solid's faces.'''
    return (
        (0, 1, 2),
        (0, 2, 3),
        (0, 3, 4),
        (0, 4, 1),
        (5, 1, 4),
        (5, 2, 1),
        (5, 3, 2),
        (5, 4, 3),
    )

def get_dodecahedron_points(center, circumradius):
    '''Return a tuple of dodecahedron vertices.'''
    # Intermediate values.
    ds = 2
    # dt1 = 2 * math.pi / 5    # Not actually used.
    dt2 = math.pi / 10
    dt3 = 3 * math.pi / 10
    dt4 = math.pi / 5
    dd1 = ds / 2 / math.sin(dt4)
    dd2 = dd1 * math.cos(dt4)
    dd3 = dd1 * math.cos(dt2)
    dd4 = dd1 * math.sin(dt2)
    d_fx = (ds * ds - (2 * dd3) * (2 * dd3) - \
            (dd1 * dd1 - dd3 * dd3 - dd4 * dd4)) / (2 * (dd4 - dd1))
    dd5 = math.sqrt(
        0.5 * (ds * ds + (2 * dd3) * (2 * dd3) -
            (dd1 - d_fx) * (dd1 - d_fx) -
            (dd4 - d_fx) * (dd4 - d_fx) - dd3 * dd3))
    d_fy = (d_fx * d_fx - dd1 * dd1 - dd5 * dd5) / (2 * dd5)
    d_ay = dd5 + d_fy
    dodecahedron_circumradius = math.sqrt(dd1 * dd1 + d_ay * d_ay)

    # Make the points.
    scale = circumradius / dodecahedron_circumradius
    a = center + scale * Vector3(dd1, d_ay, 0)
    b = center + scale * Vector3(dd4, d_ay, dd3)
    c = center + scale * Vector3(-dd2, d_ay, ds / 2)
    d = center + scale * Vector3(-dd2, d_ay, -ds / 2)
    e = center + scale * Vector3(dd4, d_ay, -dd3)
    f = center + scale * Vector3(d_fx, d_fy, 0)
    g = center + scale * Vector3(d_fx * math.sin(dt2), d_fy, d_fx * math.cos(dt2))
    h = center + scale * Vector3(-d_fx * math.sin(dt3), d_fy, d_fx * math.cos(dt3))
    i = center + scale * Vector3(-d_fx * math.sin(dt3), d_fy, -d_fx * math.cos(dt3))
    j = center + scale * Vector3(d_fx * math.sin(dt2), d_fy, -d_fx * math.cos(dt2))
    k = center + scale * Vector3(d_fx * math.sin(dt3), -d_fy, d_fx * math.cos(dt3))
    l = center + scale * Vector3(-d_fx * math.sin(dt2), -d_fy, d_fx * math.cos(dt2))
    m = center + scale * Vector3(-d_fx, -d_fy, 0)
    n = center + scale * Vector3(-d_fx * math.sin(dt2), -d_fy, -d_fx * math.cos(dt2))
    o = center + scale * Vector3(d_fx * math.sin(dt3), -d_fy, -d_fx * math.cos(dt3))
    p = center + scale * Vector3(dd2, -d_ay, ds / 2)
    q = center + scale * Vector3(-dd4, -d_ay, dd3)
    r = center + scale * Vector3(-dd1, -d_ay, 0)
    s = center + scale * Vector3(-dd4, -d_ay, -dd3)
    t = center + scale * Vector3(dd2, -d_ay, -ds / 2)
    return (a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t)

def get_dodecahedron_faces():
    '''Return a tuple holding tuples that define the solid's faces.'''
    return (
        (4, 3, 2, 1, 0),
        (0, 1, 6, 10, 5),
        (0, 5, 14, 9, 4),
        (4, 9, 13, 8, 3),
        (3, 8, 12, 7, 2),
        (2, 7, 11, 6, 1),
        (10, 15, 19, 14, 5),
        (14, 19, 18, 13, 9),
        (13, 18, 17, 12, 8),
        (12, 17, 16, 11, 7),
        (11, 16, 15, 10, 6),
        (15, 16, 17, 18, 19),
    )

def get_icosahedron_points(center, circumradius):
    '''Return a tuple of icosahedron vertices.'''
    # Intermediate values.
    s = 2
    # t1 = 2 * math.pi / 5     # Not actually used.
    t2 = math.pi / 10
    t4 = math.pi / 5
    # t3 = -3 * math.pi / 10   # Not actually used.
    r = (s / 2) / math.sin(t4)
    h0 = r * math.cos(t4)
    h1 = math.sqrt(s * s - r * r)
    h2 = math.sqrt((h0 + r) * (h0 + r) - h0 * h0)
    cx = r * math.sin(t2)
    cz = r * math.cos(t2)
    y2 = (h2 - h1) / 2
    y1 = y2 + h1
    icosahedron_circumradius = y1

    # Make the points.
    scale = circumradius / y1
    a = center + scale * Vector3(0, y1, 0)
    b = center + scale * Vector3(r, y2, 0)
    c = center + scale * Vector3(cx, y2, cz)
    d = center + scale * Vector3(-h0, y2, s / 2)
    e = center + scale * Vector3(-h0, y2, -s / 2)
    f = center + scale * Vector3(cx, y2, -cz)
    g = center + scale * Vector3(-r, -y2, 0)
    h = center + scale * Vector3(-cx, -y2, -cz)
    i = center + scale * Vector3(h0, -y2, -s / 2)
    j = center + scale * Vector3(h0, -y2, s / 2)
    k = center + scale * Vector3(-cx, -y2, cz)
    l = center + scale * Vector3(0, -y1, 0)
    return (a, b, c, d, e, f, g, h, i, j, k, l)

def get_icosahedron_faces():
    '''Return a tuple holding tuples that define the solid's faces.'''
    return (
        (0, 2, 1),
        (0, 3, 2),
        (0, 4, 3),
        (0, 5, 4),
        (0, 1, 5),
        (3, 10, 2),
        (2, 10, 9),
        (2, 9, 1),
        (1, 9, 8),
        (1, 8, 5),
        (5, 8, 7),
        (5, 7, 4),
        (4, 7, 6),
        (4, 6, 3),
        (3, 6, 10),
        (11, 9, 10),
        (11, 8, 9),
        (11, 7, 8),
        (11, 6, 7),
        (11, 10, 6),
    )


# In[12]:


# Mathy cell 3.
import math

# This cell holds methods that generate parametric surfaces.

def quadratic(x, z):
    y = 3.0 - (x * x + z * z) / 5.0
    return Vector3(x, y, z)

def splash(x, z):
    r2 = x * x + z * z
    y = 1 + 4 * math.cos(r2) / (2 + r2)
    return Vector3(x, y, z)

def strange(x, z):
    r2 = (x * x + z * z) / 4
    r = math.sqrt(r2)

    theta = math.atan2(z, x)
    y = 3 * math.exp(-r2) * math.sin(2 * math.pi * r) * math.cos(3 * theta)
    return Vector3(x, y, z)

def torus(u, v):
    major_radius = 2.5
    minor_radius = 1

    x = (major_radius + minor_radius * math.cos(v)) * math.cos(u)
    y = minor_radius * math.sin(v)
    z = (major_radius + minor_radius * math.cos(v)) * math.sin(u)
    return Vector3(x, y, z)

def mobius(u, v):
    x = (1 + v / 2 * math.cos(u / 2)) * math.sin(u)
    y = v / 2 * math.sin(u / 2)
    z = (1 + v / 2 * math.cos(u / 2)) * math.cos(u)
    scale = 2
    return Vector3(scale * x, scale * y, scale * z)

def ellipsoid(theta, phi):
    x = 3 * math.cos(theta) * math.cos(phi)
    y = 1 * math.sin(theta)
    z = 2 * math.cos(theta) * math.sin(phi)
    return Vector3(x, y, z)

def shell(u, v):
    sinu2 = math.sin(u) * math.sin(u)
    x = math.pow(1.2, v) * (sinu2 * math.cos(v))
    y = math.pow(1.2, v) * (math.sin(u) * math.cos(u))
    z = 1 - math.pow(1.2, v) * (sinu2 * math.sin(v))
    return Vector3(x, y, z)

def helix(u, v):
    x = 3 * (1 - 0.1 * math.cos(v)) * math.cos(u);
    y = 0.3 * (math.sin(v) + u / 1.7 - 5);
    z = 3 * (1 - 0.1 * math.cos(v)) * math.sin(u);
    return Vector3(x, y, z)


# In[13]:


class RtObject:
    def __init__(self, name, material):
        self.name = name
        self.material = material

    def find_hit(ray_origin, ray_direction):
        '''See where the ray hits this object.'''
        raise NotImplementedError('Method find_hit not implemented')

    # If the object is two-sided, reset its normal to point toward the camera.
    # The default implementation does nothing.
    def reset_normal(self, camera_position):
        pass


# In[14]:


class RtSphere(RtObject):
    def __init__(self, name, material, center, radius):
        super().__init__(name, material)
        self.center = center
        self.radius = radius

    def find_hit(self, ray_origin, ray_direction):
        '''See where the ray hits this object.'''
        hit_info = intersect_ray_sphere(self, ray_origin, ray_direction)
        if hit_info.is_hit():
            hit_info.normal = self.find_normal(hit_info.point)
        return hit_info

    def find_normal(self, point):
        '''Return the unit normal at a specific point.'''
        normal = point - self.center

        if self.material.bump_map is not None:
            # Apply the bump map.
            normal = self.apply_bump_map(point, normal)
        elif self.material.normal_map is not None:
            # Apply the normal map.
            normal = self.apply_normal_map(point, normal)

        return normal.normalize()

    def point_to_bump_coords(self, point):
        '''Return the point in bump map coordinate space (plus phi).'''
        # The point parameter is the vector from the sphere's center to the point.
        # Convert the point to spherical coordinates.
        rho, theta, phi = cartesian_to_spherical(point)

        # Scale theta and phi to values between 0 and 1.
        theta = (theta + math.pi) / (2 * math.pi)
        phi = phi / math.pi

        # Scale theta and phi to u and v in bump map coordinates.
        theta *= self.material.bump_map_size[0] - 1
        phi   *= self.material.bump_map_size[1] - 1

        return Vector3(rho, theta, phi)

    def bump_coords_to_point(self, point):
        '''Return the point converted back into Cartesian coordinates.'''
        # The point parameter is the vector from the sphere's center to the point.
        # Convert the point to spherical coordinates.
        rho, theta, phi = point

        # Scale theta and phi to values between 0 and 1.
        theta /= self.material.bump_map_size[0] - 1
        phi   /= self.material.bump_map_size[1] - 1

        # Scale so -pi <= theta <= pi.
        theta = theta * 2 * math.pi - math.pi

        # Scale so 0 <= phi <= pi.
        phi = phi * math.pi

        return spherical_to_cartesian(Vector3(rho, theta, phi))

    def apply_bump_map(self, point_a, normal):
        '''Return an adjusted normal vector.'''
        # If the point doesn't exist, do nothing.
        if math.isinf(point_a.x) or math.isinf(point_a.y) or math.isinf(point_a.z):
            return normal

        # Convert the point to bump map coordinates.
        coords_a = self.point_to_bump_coords(point_a - self.center)

        # Get texture coordinates of nearby points.
        u_b = (coords_a.y + 1) % self.material.bump_map_size[0]
        coords_b = Vector3(coords_a.x, u_b, coords_a.z)

        v_c = (coords_a.z + 1) % self.material.bump_map_size[1]
        coords_c = Vector3(coords_a.x, coords_a.y, v_c)

        # Get heights for all three points.
        coords_a.x += self.material.bump_map_height(coords_a.y, coords_a.z)
        coords_b.x += self.material.bump_map_height(coords_b.y, coords_b.z)
        coords_c.x += self.material.bump_map_height(coords_c.y, coords_c.z)

        # Convert the points to Cartesian coordinates.
        point_a = self.bump_coords_to_point(coords_a)
        point_b = self.bump_coords_to_point(coords_b)
        point_c = self.bump_coords_to_point(coords_c)

        # Calculate vectors ab and ac.
        ab = point_b - point_a
        ac = point_c - point_a

        # Some bump map values and normals may give impossible results here.
        try:
            # Take the cross product so get the adjusted normal vector.
            cross = -ab.cross(ac)

            # Normalize and return.
            return cross.normalize()
        except:
            return normal

    def apply_normal_map(self, point, normal):
        '''Return the point's normal from the normal map.'''
        # The point parameter is the vector from the sphere's center to the point.
        if math.isinf(point.x) or math.isinf(point.y) or math.isinf(point.z):
            return normal

        # Convert the point to spherical coordinates.
        rho, theta, phi = cartesian_to_spherical(point - self.center)

        # Scale theta and phi to values between 0 and 1.
        theta = (theta + math.pi) / (2 * math.pi)
        phi = phi / math.pi

        # Scale theta and phi to u and v in normal map coordinates.
        theta *= self.material.normal_map_size[0] - 1
        phi   *= self.material.normal_map_size[1] - 1

        # Get the normal map value.
        r, g, b = self.material.normal_map[theta, phi]

        # Convert the rgb values from a 0 to 255 scale to a -1 to 1 scale.
        x = r / 255 * 2 - 1
        y = g / 255 * 2 - 1
        z = b / 255 * 2 - 1

        # Return the normal vector.
        return Vector3(x, y, z)


# In[15]:


import math

class RtTexturedSphere(RtSphere):
    def color_at_point(self, hit_info):
        '''Return the texture color at this point.'''
        # Get the vector from the sphere's center to the hit point.
        vector = hit_info.point - self.center
        r, theta, z = cartesian_to_cylindrical(vector)

        # Convert (theta, z) into texture coordinates in the range 0.0 to 1.0.
        coords_u = (theta + math.pi) / (2 * math.pi)
        coords_v = (z + self.radius) / (2 * self.radius)

        # Scale to fit the texture.
        x = coords_u * (self.material.texture_size[0] - 1)
        y = coords_v * (self.material.texture_size[1] - 1)

        # Use bilinear interpolation to find the texture's color at this point.
        return interpolate_color(self.material.texture, self.material.texture_size, x, y)


# In[16]:


class RtPlane(RtObject):
    def __init__(self, name, material, origin, normal):
        super().__init__(name, material)
        self.origin = origin
        self.normal = normal.normalize()

    def find_hit(self, ray_origin, ray_direction):
        return intersect_ray_plane(self, ray_origin, ray_direction)

    # Planes are two-sided. Reset the normal to point toward the camera.
    def reset_normal(self, camera_position):
        # If the normal points more or less away from the camera, reverse it.
        camera_vector = camera_position - self.origin
        if camera_vector.dot(self.normal) < 0:
            self.normal = -self.normal


# In[17]:


class RtDisk(RtPlane):
    def __init__(self, name, material, origin, normal, radius):
        super().__init__(name, material, origin, normal)
        self.radius = radius

    def find_hit(self, ray_origin, ray_direction):
        # Find the plane's hit.
        hit_info = intersect_ray_plane(self, ray_origin, ray_direction)

        # If there is no intersection, we're done.
        if hit_info.is_miss(): return hit_info

        # See if the point is within the radius.
        distance = (self.origin - hit_info.point).length()
        if distance <= self.radius:
            # Return the hit info.
            return hit_info
        else:
            # Return a miss.
            return RtHitInfo(self)

    # Inherit reset_normal from RtPlane.


# In[18]:


class RtParallelogram(RtPlane):
    def __init__(self, name, material, a, ab, ac):
        super().__init__(name, material, a, ab.cross(ac))
        self.a = a
        self.ab = ab
        self.ac = ac

    def find_hit(self, ray_origin, ray_direction):
        '''See where the ray hits this object.'''
        return intersect_ray_parallelogram(self, ray_origin, ray_direction)


# In[19]:


class RtTriangle(RtParallelogram):
    # Constructor is inherited from RtParallelogram.

    @staticmethod
    def from_points(name, material, a, b, c):
        '''Create a triangle from three points.'''
        return RtTriangle(name, material, a, b - a, c - a)

    def find_hit(self, ray_origin, ray_direction):
        '''See where the ray hits this object.'''
        return intersect_ray_triangle(self, ray_origin, ray_direction)


# In[20]:


class RtCheckerboard(RtParallelogram):
    def __init__(self, name, material, a, ab, ac, num_ab, num_ac, parity):
        super().__init__(name, material, a, ab, ac)
        self.num_ab = num_ab
        self.num_ac = num_ac
        self.parity = parity

    def find_hit(self, ray_origin, ray_direction):
        '''See where the ray hits this object.'''
        return intersect_ray_checkerboard(self, ray_origin, ray_direction)


# In[21]:


from pygame import Vector3

class RtMesh(RtObject):
    def __init__(self, name, material):
        super().__init__(name, material)

        # Make empty collections to hold points, indices, etc. 
        self.vertices = []
        self.triangle_indices = []
        self.triangle_normals = []
        self.ab_vectors = []
        self.ac_vectors = []
        self.shared_point_indices = {}
        self.is_smooth = False
        self.vertex_normals = []
        self.texture_coords = []

    def get_vertex_index(self, point, reuse_points=False, coords=None):
        '''Find or create a vertex.'''
        # Compose the point's key string.
        key = str(point)

        # See if we should reuse points.
        if reuse_points:
            # If the vertex already exists, return its index.
            if key in self.shared_point_indices:
                return self.shared_point_indices[key]

        # Either we should not reuse points or the vertex doesn't yet exist.
        # Create a new vertex.
        self.vertices.append(point)
        vertex_index = len(self.vertices) - 1

        # Save the texture coordinates.
        if coords is not None:
            self.texture_coords.append(coords)

        # If we're sharing this vertex, add it to the dictionary.
        if reuse_points:
            self.shared_point_indices[key] = vertex_index

        # Return the vertex's index.
        return vertex_index

    def add_triangle(self, a, b, c, reuse_points=False,
            coords_a=None, coords_b=None, coords_c=None):
        '''Add a triangle to the mesh.'''
        # Find the triangle's vectors and normal.
        ab = b - a
        ac = c - a
        try:
            normal = ab.cross(ac).normalize()
        except:
            # If anything went wrong, bail out.
            return

        # Get the points' indices.
        index_a = self.get_vertex_index(a, reuse_points, coords_a)
        index_b = self.get_vertex_index(b, reuse_points, coords_b)
        index_c = self.get_vertex_index(c, reuse_points, coords_c)

        # Make the triangle. (The entry is a tuple of point indexes.)
        self.triangle_indices.append( (index_a, index_b, index_c) )
        self.ab_vectors.append(ab)
        self.ac_vectors.append(ac)
        self.triangle_normals.append(normal)

    def add_polygon(self, points, reuse_points=False, coords_list=None):
        '''Add a polygon as a fan of triangles.'''
        if coords_list is None:
            for i in range(2, len(points)):
                self.add_triangle(points[0], points[i - 1], points[i], reuse_points)
        else:
            for i in range(2, len(points)):
                self.add_triangle(points[0], points[i - 1], points[i],
                    reuse_points,
                    coords_list[0], coords_list[i - 1], coords_list[i])

    def reset_normal(self, camera_position):
        '''Reset the mesh's triangles to point toward the camera.'''
        # Loop through the triangles and reset their normals.
        for triangle_num in range(len(self.triangle_indices)):
            # Get the triangle's first vertex and vectors.
            a = self.vertices[self.triangle_indices[triangle_num][0]]
            camera_vector = camera_position - a
            if self.triangle_normals[triangle_num].dot(camera_vector) < 0:
                self.triangle_normals[triangle_num] = \
                    -self.triangle_normals[triangle_num]

    def find_hit(self, ray_origin, ray_direction):
        # Keep track of the intersection closest to the ray's origin.
        best_hit_info = RtHitInfo(self)

        # Loop through the triangles to see if one is hit.
        for triangle_num in range(len(self.triangle_indices)):
            # See if there is an intersection.
            hit_info = intersect_ray_mesh_triangle(self,
                triangle_num, ray_origin, ray_direction)

            # If this is closer to the ray's origin than
            # best_hit_info, update best_hit_info.
            if hit_info.t < best_hit_info.t:
                best_hit_info = hit_info

        # Return the best RtHitInfo.
        return best_hit_info

    def add_vector_polygon(self, center, v1, v2, num_sides):
        '''Add triangles to make a polygon centered at center and in the plane defined by vectors v1 and v2.'''
        self.add_polygon(make_polygon_points(center, v1, v2, num_sides))

    def add_pyramid(self, apex, center, v1, v2, num_sides):
        '''Add triangles to make a pyramid with the given apex and base defined by the center, v1, and v2 parameters.'''
        # Get the base's points.
        points = make_polygon_points(center, v1, v2, num_sides)

        # Build the sides.
        for i in range(num_sides):
            self.add_triangle(apex, points[i], points[(i + 1) % num_sides])

        # Build the base.
        self.add_polygon(points)

    def add_solid(self, center, circumradius, get_points, get_faces, reuse_points=False, get_texture_coords=None):
        '''Make triangles for the solid.'''
        # Get the points.
        points = get_points(center, circumradius)

        # Get the faces.
        faces = get_faces()

        if get_texture_coords is None:
            # Make the faces.
            for face in faces:
                # Make a list of this face's points.
                face_points = [points[index] for index in face]

                # Make the face's triangles.
                self.add_polygon(face_points, reuse_points)
        else:
            # Get the solid's texture coordinates.
            coords_list = get_texture_coords()

            # Make the faces.
            for face, coords in zip(faces, coords_list):
                # Make a list of this face's points.
                face_points = [points[index] for index in face]
    
                # Make the face's triangles.
                self.add_polygon(face_points, reuse_points, coords)

    def add_tetrahedron(self, center, circumradius, reuse_points=False):
        self.add_solid(center, circumradius, get_tetrahedron_points, get_tetrahedron_faces, reuse_points)

    def add_cube(self, center, circumradius, reuse_points=False):
        self.add_solid(center, circumradius, get_cube_points, get_cube_faces, reuse_points)

    def add_textured_cube(self, center, circumradius, reuse_points=False):
        self.add_solid(center, circumradius, get_cube_points, get_cube_faces, reuse_points, get_cube_texture_coords)

    def add_octahedron(self, center, circumradius, reuse_points=False):
        self.add_solid(center, circumradius, get_octahedron_points, get_octahedron_faces, reuse_points)

    def add_dodecahedron(self, center, circumradius, reuse_points=False):
        self.add_solid(center, circumradius, get_dodecahedron_points, get_dodecahedron_faces, reuse_points)

    def add_icosahedron(self, center, circumradius, reuse_points=False):
        self.add_solid(center, circumradius, get_icosahedron_points, get_icosahedron_faces, reuse_points)

    def add_surface(self, f, xmin, xmax, num_x, zmin, zmax, num_z, reuse_points=False):
        '''Add triangles to draw the surface defined by y = f(x, z).'''
        dx = (xmax - xmin) / num_x
        dz = (zmax - zmin) / num_z
        x = xmin
        for ix in range(num_x):
            z = zmin
            for iz in range(num_z):
                p1 = f(x, z)
                p2 = f(x, z + dz)
                p3 = f(x + dx, z + dz)
                p4 = f(x + dx, z)
                self.add_polygon((p1, p2, p3, p4), reuse_points)
                z += dz
            x += dx

    def add_textured_surface(self, f, xmin, xmax, num_x, zmin, zmax, num_z,
            reuse_points=False, umin=0, umax=1, vmin=0, vmax=1):
        '''Add triangles to draw the surface defined by y = f(x, z).'''
        dx = (xmax - xmin) / num_x
        dz = (zmax - zmin) / num_z
        du = (umax - umin) / num_x
        dv = (vmax - vmin) / num_z
        dx = (xmax - xmin) / num_x
        dz = (zmax - zmin) / num_z
        x = xmin
        u = umin
        for ix in range(num_x):
            z = zmin
            v = vmin
            for iz in range(num_z):
                p1 = f(x, z)
                p2 = f(x, z + dz)
                p3 = f(x + dx, z + dz)
                p4 = f(x + dx, z)
                coords1 = Vector2(u, v)
                coords2 = Vector2(u, v + dv)
                coords3 = Vector2(u + du, v + dv)
                coords4 = Vector2(u + du, v)
                self.add_polygon((p1, p2, p3, p4), reuse_points,
                    (coords1, coords2, coords3, coords4))
                z += dz
                v += dv
            x += dx
            u += du

    def find_smooth_normal(self, point, triangle_num, v):
        '''Return the smoothed unit normal for this intersection.'''
        # Find the triangle's vertices and normals.
        index_a = self.triangle_indices[triangle_num][0]
        index_b = self.triangle_indices[triangle_num][1]
        index_c = self.triangle_indices[triangle_num][2]
        point_a = self.vertices[index_a]
        point_b = self.vertices[index_b]
        point_c = self.vertices[index_c]
        na = self.vertex_normals[index_a]
        nb = self.vertex_normals[index_b]
        nc = self.vertex_normals[index_c]

        # Interpolate.
        normal = interpolate(point, point_a, point_b, point_c, na, nb, nc, v)

        # Return the normalized vector.
        return normal.normalize()

    def smooth(self):
        '''Calculate vertex normals for smoothing.'''
        # Calculate the vertex normals.
        num_vertices = len(self.vertices)
        self.vertex_normals = [Vector3() for i in range(num_vertices)]

        # Keep track of the number of triangles that uses each vertex.
        use_count = [0 for i in range(num_vertices)]

        # Loop through the triangles.
        for tri in range(len(self.triangle_indices)):
            # Process the triangle's vertices.
            for p in range(3):
                # Find this vertex's index.
                point_index = self.triangle_indices[tri][p]

                # Add the triangle's normal to this vertex's normal total.
                self.vertex_normals[point_index] += self.triangle_normals[tri]
                use_count[point_index] += 1

        # Average each vertex's total normal and normalize.
        for i in range(num_vertices):
            self.vertex_normals[i] /= use_count[i]
            self.vertex_normals[i].normalize()

    def add_triangle_from_vertices(self, index_a, index_b, index_c, reuse_points=False):
        '''Add a triangle to the mesh given its vertex indices.'''
        # Get the points.
        a = self.vertices[index_a]
        b = self.vertices[index_b]
        c = self.vertices[index_c]

        # Find the triangle's vectors and normal.
        ab = b - a
        ac = c - a
        try:
            normal = ab.cross(ac).normalize()
        except:
            # If anything went wrong, bail out.
            return

        # Make the triangle. (The entry is a tuple of point indexes.)
        self.triangle_indices.append( (index_a, index_b, index_c) )
        self.ab_vectors.append(ab)
        self.ac_vectors.append(ac)
        self.triangle_normals.append(normal)

    # http://www.holmes3d.net/graphics/teapot/teapot.off
    def add_teapot(self, filename, offset=Vector3(0, 0, 0), scale=1, use_phong_shading=False):
        ''''Add the Utah teapot.'''
        self.is_smooth = use_phong_shading

        with open(filename) as file:
            # Ignore OFF.
            line = file.readline().strip()
            if line == 'OFF': line = file.readline().strip()

            # Get the number of vertices.
            fields = line.split()
            num_vertices = int(fields[0])
            num_faces = int(fields[1])
            num_edges = int(fields[2])

            # Get the vertices.
            for i in range(num_vertices):
                line = file.readline().strip()
                fields = line.split()
                # Switch Y and Z coordinates.
                point = scale * (offset + Vector3(float(fields[0]), float(fields[2]), float(fields[1])))
                self.vertices.append(point)

            # Get the faces.
            for i in range(num_faces):
                line = file.readline().strip()
                indices = [int(x) for x in line.split()]
                for j in range(3, len(indices)):
                    self.add_triangle_from_vertices(indices[1], indices[j - 1], indices[j])

        if False:
            # Find the teapot's bounds.
            xmin = min(p.x for p in self.vertices)
            xmax = max(p.x for p in self.vertices)
            xave = (xmin + xmax) / 2
            print(f'X min: {xmin}')
            print(f'X max: {xmax}')
            print(f'X ave: {xave}')

            ymin = min(p.y for p in self.vertices)
            ymax = max(p.y for p in self.vertices)
            yave = (ymin + ymax) / 2
            print(f'Y min: {ymin}')
            print(f'Y max: {ymax}')
            print(f'Y ave: {yave}')

            zmin = min(p.z for p in self.vertices)
            zmax = max(p.z for p in self.vertices)
            zave = (zmin + zmax) / 2
            print(f'Z min: {zmin}')
            print(f'Z max: {zmax}')
            print(f'Z ave: {zave}')

            # Display an appropriate offset vector.
            print(f'offset = Vector3(-{xave}, -{yave}, -{zave})')

    def color_at_point(self, hit_info):
        '''Use bilinear interpolation to return the color at this point.'''
        # Get the interpolated texture coordinates.
        coords = hit_info.coords

        # Scale to fit the texture.
        x = coords.x * (self.material.texture_size[0] - 1)
        y = coords.y * (self.material.texture_size[1] - 1)

        # Use bilinear interpolation to find the texture's color at this point.
        return interpolate_color(self.material.texture, self.material.texture_size, x, y)


# In[22]:


class RtCylinder(RtObject):
    def __init__(self, name, material, origin, axis, radius, length):
        super().__init__(name, material)
        self.origin = origin
        self.axis = axis.normalize()
        self.radius = radius
        self.length = length

    def find_hit(self, ray_origin, ray_direction):
        '''See where the ray hits this object.'''
        return intersect_ray_cylinder(self, ray_origin, ray_direction)

    def find_normal(self, point):
        '''Find the cylinder's normal at the point.'''
        L = (point - self.origin).dot(self.axis)
        closest = self.origin + self.axis * L
    
        # The normal points from closest to point.
        normal = point - closest
        return normal.normalize()

    @staticmethod
    def from_endpoints(name, material, a, b, radius):
        axis = b - a
        length = axis.length()
        return RtCylinder(name, material, a, axis, radius, length)


# In[23]:


from PIL import Image
import time

class RtScene:
    def __init__(self, background_color, camera):
        self.background_color = background_color
        self.camera = camera
        self.objects = []
        self.lights = []
        self.refractive_index = 1.0

    def find_hit(self, ray_origin, ray_direction):
        '''See what the ray hits.'''
        # Assume we hit nothing.
        best_hit_info = RtHitInfo()

        # Check each object. Find the hit with smallest T value.
        for object in self.objects:
            hit_info = object.find_hit(ray_origin, ray_direction)
            if hit_info.t < best_hit_info.t:
                best_hit_info = hit_info
        return best_hit_info

    def find_color(self, ray_origin, ray_direction, depth, max_depth):
        '''Trace the indicated ray, see what we hit, and determine the hit color.'''
        # See what we hit.
        hit_info = self.find_hit(ray_origin, ray_direction)

        # If we hit nothing, return the background color.
        if hit_info.is_miss(): return self.background_color

        # Examine the lights.
        total_color = RtColor(0, 0, 0)
        for light in self.lights:
            total_color += light.find_hit_color(self, hit_info)

        # If the object is reflective, add in a reflective component.
        if (depth < max_depth) and (hit_info.object.material.reflectivity > 0.0):
            mirror_vector = find_mirror_vector(ray_direction, hit_info.normal)
            new_color = self.find_color(hit_info.point, mirror_vector, depth + 1, max_depth)
            total_color += new_color * hit_info.object.material.reflectivity

        # If the object is at least partially transparent, add in a refractive component.
        if (depth < max_depth) and (hit_info.object.material.refractive_intensity > 0.0):
            # Initially assume we're entering the object.
            ni = self.refractive_index
            nt = hit_info.object.material.refractive_index

            # Find the refraction vector.
            refractive_vector = find_refraction_vector(
                ray_direction, hit_info.normal, ni, nt)

            # Ignore if the vector has zero length, indicating internal reflection.
            if refractive_vector is not None:
                # Follow the refraction vector.
                new_color = self.find_color(hit_info.point, refractive_vector,
                    depth + 1, max_depth)

                # Add the new color contribution.
                total_color += new_color * hit_info.object.material.refractive_intensity

        # Return the total color from all lights.
        return total_color

    def make_image(self, width, height, max_depth, num_images=1, aperture_radius=0):
        '''Make an Image holding the ray traced scene.'''
        start_time = time.time()

        # Make the Image.
        image = Image.new(mode='RGB', size=(width, height))
        pixels = image.load()

        # Reset the normals of two-sided objects.
        for object in self.objects:
            object.reset_normal(self.camera.position)

        # Smooth the meshes that use Phong shading.
        for object in self.objects:
            if type(object) is RtMesh and object.is_smooth:
                object.smooth()

        # World coordinates per pixel.
        scale = self.camera.screen_width / width

        # Get one-pixel long vectors in the right and down directions.
        right = self.camera.direction.cross(self.camera.up)
        down = self.camera.direction.cross(right)
        right.scale_to_length(scale)
        down.scale_to_length(scale)

        # Find the center of the screen.
        camera_to_screen = Vector3(self.camera.direction)
        camera_to_screen.scale_to_length(self.camera.screen_distance)
        center = self.camera.position + camera_to_screen

        # Find the upper left corner of the screen.
        ul = center - right * width / 2 - down * height / 2

        # Get aperature vectors.
        aperture_x = right.normalize()
        aperture_y = down.normalize()

        # Loop over the screen's pixels.
        for y in range(height):
            print('.', end='')
            if y % 100 == 99: print()

            point = ul + y * down
            for x in range(width):
                # Add up the colors from each randomized ray.
                color = RtColor(0, 0, 0)

                for image_num in range(num_images):
                    # Randomize the camera's position.
                    camera_position = self.camera.position + \
                        random_disk_vector(aperture_x, aperture_y, aperture_radius)

                    # Trace a ray from the randomized camera position through this point.
                    ray_direction = (point - camera_position).normalize()
                    color += self.find_color(camera_position, ray_direction, 0, max_depth)

                # Divide the summed color by the number of randomized rays.
                color *= 1 / num_images

                pixels[x, y] = color.to_rgb_tuple()
                point += right

        if height % 100 != 0: print()
        end_time = time.time()
        elapsed_time = round(end_time - start_time)
        print(f'{elapsed_time} seconds')

        # Return the Image.
        return image


# In[44]:


import tkinter as tk
from tkinter import ttk
from pygame import Vector3
from PIL import ImageTk, Image
import random

class RtApp:
    def __init__(self):
        # Make the interface. (We need to do this before drawing
        # so we can get the Canvas control's dimensions.)
        self.window = tk.Tk()
        self.window.title('RayTracer23')
        self.window.geometry('625x425')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)

        # Add a button area.
        button_frame_width = 189
        button_frame = tk.Frame(self.window, width=button_frame_width,
            borderwidth=2, relief=tk.SUNKEN)
        button_frame.pack(padx=4, pady=4, side=tk.LEFT, fill=tk.Y, expand=False)
        button_frame.pack_propagate(0) # Prevent the frame from shrinking to fit.

        # Set some layout constants.
        label_width = 150
        label_height = 24
        entry_width = 5

        # Make the image dimension textboxes.
        self.image_width = self.make_entry(button_frame,
            label_width, label_height, 'Image Width:', entry_width, '400', tk.RIGHT)
        self.image_height = self.make_entry(button_frame,
            label_width, label_height, 'Image Height:', entry_width, '400', tk.RIGHT)

        # Make the camera position textboxes.
        self.camera_x = self.make_entry(button_frame,
            label_width, label_height, 'Camera X:', entry_width, '4', tk.RIGHT)
        self.camera_y = self.make_entry(button_frame,
            label_width, label_height, 'Camera Y:', entry_width, '3', tk.RIGHT)
        self.camera_z = self.make_entry(button_frame,
            label_width, label_height, 'Camera Z:', entry_width, '6', tk.RIGHT)

        # Make the screen parameter textboxes.
        self.viewing_angle = self.make_entry(button_frame,
            label_width, label_height, 'Viewing Angle:', entry_width, '60', tk.RIGHT)

        # Make the Max Depth textbox.
        self.max_depth = self.make_entry(button_frame,
            label_width, label_height, 'Max Depth:', entry_width, '4', tk.RIGHT)

        # Make the Super Samples textbox.
        self.super_samples = self.make_entry(button_frame,
            label_width, label_height, 'Super Samples:', entry_width, '1', tk.RIGHT)

        # Add the Draw button.
        self.draw_button = tk.Button(button_frame, text='Draw', width=8,
            command=self.draw_callback)
        self.draw_button.pack(padx=4, pady=5, side=tk.TOP, fill=tk.NONE, expand=False)

        # Checkboxes.
        self.draw_origin = tk.IntVar(self.window, 0)
        draw_origin_button = tk.Checkbutton(button_frame, text='Origin',
            variable=self.draw_origin, onvalue=1, offvalue=0)
        draw_origin_button.pack(padx=4, pady=5, side=tk.TOP, fill=tk.NONE, expand=False, anchor='w')

        self.draw_ground = tk.IntVar(self.window, 1)
        draw_ground_button = tk.Checkbutton(button_frame, text='Ground',
            variable=self.draw_ground, onvalue=1, offvalue=0)
        draw_ground_button.pack(padx=4, pady=5, side=tk.TOP, fill=tk.NONE, expand=False, anchor='w')

        self.draw_tubes = tk.IntVar(self.window, 1)
        draw_tubes_button = tk.Checkbutton(button_frame, text='Tubes',
            variable=self.draw_tubes, onvalue=1, offvalue=0)
        draw_tubes_button.pack(padx=4, pady=5, side=tk.TOP, fill=tk.NONE, expand=False, anchor='w')

        # Fire the Draw button when you press Return.
        self.window.bind('<Return>', lambda e: self.draw_button.invoke())

        # Add a Canvas.
        self.canvas = tk.Canvas(self.window, bg='white',
            borderwidth=2, relief=tk.SUNKEN)
        self.canvas.pack(padx=(0,4), pady=2, side=tk.RIGHT, fill=tk.BOTH, expand=True)

        # Run the tkinter main loop.
        self.window.mainloop()

    def kill_callback(self):
        '''Destroy the tkinter window.'''
        self.window.destroy()

    def draw_callback(self):
        '''The user clicked the Draw button. Draw the scene.'''
        self.window.config(cursor='watch')
        self.window.update()

        # Define and draw the scene.
        self.define_scene()
        self.draw_scene()

        # Save the image.
        self.image.save('output.png')

        # Remove any previous children on the Canvas widget.
        for child in self.canvas.winfo_children():
            child.destroy()

        # Display the image on the Canvas.
        self.photo_image = ImageTk.PhotoImage(self.image)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.photo_image)

        self.window.config(cursor='')

    # Define the scene.
    def define_scene(self):
        # Make the camera.
        camera_x = float(self.camera_x.get())
        camera_y = float(self.camera_y.get())
        camera_z = float(self.camera_z.get())
        position = Vector3(camera_x, camera_y, camera_z)
        direction = -position
        up = Vector3(0, 1, 0)
        distance = direction.length()
        camera = RtCamera(position, direction, up, distance)
        viewing_angle = float(self.viewing_angle.get())
        camera.viewing_angle = viewing_angle

        # Make the scene.
        self.scene = RtScene(RtColor.BLACK, camera)

        # Make lights.
        dim_gray = RtColor(0.3, 0.3, 0.3)
        self.scene.lights.append(RtAmbientLight(RtColor.DARK_GRAY))
        self.scene.lights.append(RtDirectionalLight(RtColor.GRAY, Vector3(1, -1, -1)))

        if self.draw_origin.get():
            # Draw the origin.
            origin_material = RtMaterial(RtColor.YELLOW)
            self.scene.objects.append(RtSphere('origin', origin_material, Vector3(), 0.2))

        if self.draw_ground.get():
            # Draw the ground.
            material = RtMaterial(RtColor.ORANGE)
            origin = Vector3(0, -3, 0)
            normal = Vector3(0, 1, 0)
            self.scene.objects.append(RtPlane('ground', material, origin, normal))

        if self.draw_tubes.get():
            # Draw the tubes.
            radius = 0.4
            material = RtMaterial(
                ambient_color=RtColor.GREEN,
                diffuse_color=RtColor.GREEN,
                specular_intensity=0.4, specular_power=30,
                reflectivity=0.1)
            points = [
                Vector3(2, -2, 1),
                Vector3(2, -2, -1),
                Vector3(2, 1, -1),
                Vector3(-2, 1, -1),
                Vector3(-2, 1, 2),
                Vector3(-2, 0, 2),
                Vector3(1, 0, 2),
                Vector3(1, 0, -2),
                Vector3(1, 2, -2),
                Vector3(-1, 2, -2),
                Vector3(-1, 2, 1),
                Vector3(-1, -2, 1),
            ]
            for i in range(len(points)):
                j = (i + 1) % len(points)
                self.scene.objects.append(
                    RtCylinder.from_endpoints(
                        'cylinder', material,
                        points[i], points[j], radius))
                self.scene.objects.append(
                    RtSphere('sphere', material, points[i], radius))

    # Draw the scene.
    def draw_scene(self):
        # Get the Canvas widget's size.
        width = int(self.image_width.get())
        height = int(self.image_height.get())
        max_depth = int(self.max_depth.get())
        scale = int(self.super_samples.get())

        global interpolate_texture
        interpolate_texture = True

        num_images = 1
        aperture_radius = 0

        # Make the enlarged ray traced image.
        image = self.scene.make_image(scale * width, scale * height, max_depth)

        # Scale the image to its desired size.
        self.image = image.resize((width, height), Image.Resampling.LANCZOS)

    def make_sphere_axes(self, tick_spacing, num_ticks, sphere_radius):
        '''Draw small spheres along the axes.'''
        # Make the axis materials.
        x_material = RtMaterial(RtColor.RED)
        y_material = RtMaterial(RtColor.GREEN)
        z_material = RtMaterial(RtColor.BLUE)
    
        # Make the spheres.
        for i in range(1, num_ticks + 1):
            coordinate = i * tick_spacing
            x_sphere = RtSphere(f'X = {coordinate}', x_material,
                Vector3(coordinate, 0, 0), sphere_radius)
            y_sphere = RtSphere(f'Y = {coordinate}', y_material,
                Vector3(0, coordinate, 0), sphere_radius)
            z_sphere = RtSphere(f'Z = {coordinate}', z_material,
                Vector3(0, 0, coordinate), sphere_radius)
            self.scene.objects.append(x_sphere)
            self.scene.objects.append(y_sphere)
            self.scene.objects.append(z_sphere)
    
        origin_material = RtMaterial(RtColor.WHITE)
        origin_sphere = RtSphere('Origin', origin_material,
            Vector3(0, 0, 0), sphere_radius)
        self.scene.objects.append(origin_sphere)

    def make_entry(self, parent, frame_width, frame_height, label_text,
            entry_width, entry_text, entry_justification):
        '''Make Label and Entry widgets. Return the Entry widget.'''
        # Make a Frame to hold everything.
        frame = tk.Frame(parent)
        frame.pack(side=tk.TOP, fill=tk.X, expand=False)

        # Make a Frame to hold the Label.
        label_frame = tk.Frame(frame, width=frame_width, height=frame_height)
        label_frame.pack(side=tk.LEFT, fill=tk.NONE, expand=False, anchor='w')
        label_frame.pack_propagate(0)

        # Make the Label inside label_frame.
        label = tk.Label(label_frame, text=label_text)
        label.pack(side=tk.LEFT, fill=tk.NONE, expand=False, anchor='w')

        # Make the Entry inside frame.
        entry = tk.Entry(frame, width=entry_width, justify=entry_justification)
        entry.insert(tk.END, entry_text)
        entry.pack(side=tk.LEFT, fill=tk.NONE, expand=True, anchor='w')

        return entry

    def make_combo(self, parent, frame_width, frame_height, label_text,
            combo_width, combo_values, combo_justification):
        '''Make Label and Combobox widgets. Return the Combobox widget.'''
        # Make a Frame to hold everything.
        frame = tk.Frame(parent)
        frame.pack(side=tk.TOP, fill=tk.X, expand=False)

        # Make a Frame to hold the Label.
        label_frame = tk.Frame(frame, width=frame_width, height=frame_height)
        label_frame.pack(side=tk.LEFT, fill=tk.NONE, expand=False, anchor='w')
        label_frame.pack_propagate(0)

        # Make the Label inside label_frame.
        label = tk.Label(label_frame, text=label_text)
        label.pack(side=tk.LEFT, fill=tk.NONE, expand=False, anchor='w')

        # Make the Combobox inside frame.
        combo = ttk.Combobox(frame, width=combo_width,
            justify=combo_justification, values=combo_values)
        combo.pack(side=tk.LEFT, fill=tk.NONE, expand=True, anchor='w')

        return combo


# In[45]:


app = RtApp()
app = None


# In[ ]:




